/*
 * #%L
 * JavaHg
 * %%
 * Copyright (C) 2011 aragost Trifork ag
 * %%
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * #L%
 */
package com.aragost.javahg;

import java.io.IOException;
import java.util.List;

import org.joda.time.DateTime;

import com.aragost.javahg.commands.LogCommand;
import com.aragost.javahg.internals.HgInputStream;
import com.aragost.javahg.internals.RuntimeIOException;
import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.Lists;

public class Changeset {

    public static final String CHANGESET_TEMPLATE = "\\0{node}{rev}\\n{author}\\n{date|hgdate}\\n{branch}\\n{parents}{desc}";

    public static final String NULL_ID = "0000000000000000000000000000000000000000";

    private final String node;
    private final Repository repository;
    protected ChangesetData data;

    /**
     * Use {@link Repository#changeSet(String)} to create ChangeSet's
     * 
     * @param repository
     */
    public Changeset(Repository repository, String node) {
        this.repository = repository;
        this.node = node;
    }

    private static Changeset createFromInputStream(Repository repository, HgInputStream in) throws IOException {
        byte[] node = in.next(40);
        if (node == null) {
            return null;
        }
        String nodeString = new String(node);
        int revision = in.revisionUpTo('\n');
        Changeset cset = repository.changeSet(nodeString);
        String user = in.textUpTo('\n');
        DateTime timestamp = in.dataTimeUpTo('\n');
        String branch = in.textUpTo('\n');
        in.upTo(':');
        String p1 = new String(in.next(40));
        in.upTo(':');
        String p2 = new String(in.next(40));
        in.read(); // skip space part of {parents}
        Changeset parent1 = repository.changeSet(p1);
        Changeset parent2 = repository.changeSet(p2);
        String message = in.textUpTo('\0');

        ChangesetData data = cset.data;
        if (data == null) {
            data = new ChangesetData(revision, nodeString, user, timestamp, branch, parent1, parent2, message);
            cset.data = data;
        }
        return cset;
    }

    public static List<Changeset> readListFromStream(Repository repository, HgInputStream in) throws IOException {
        List<Changeset> changesets = Lists.newArrayList();
        in.upTo('\0');
        while (true) {
            Changeset cset = Changeset.createFromInputStream(repository, in);
            if (cset == null) {
                break;
            }
            changesets.add(cset);
        }
        return changesets;
    }

    public String getNode() {
        return this.node;
    }

    public int getRevision() {
        ensureAllDataLoaded();
        return this.data.revision;
    }

    public String getUser() {
        ensureAllDataLoaded();
        return this.data.user;
    }

    public DateTime getTimestamp() {
        ensureAllDataLoaded();
        return this.data.timestamp;
    }

    public String getBranch() {
        ensureAllDataLoaded();
        return this.data.branch;
    }

    public Changeset getParent1() {
        ensureAllDataLoaded();
        return this.data.parent1;
    }

    public Changeset getParent2() {
        ensureAllDataLoaded();
        return this.data.parent2;
    }

    public String getMessage() {
        ensureAllDataLoaded();
        return this.data.message;
    }

    private void ensureAllDataLoaded() {
        if (this.data != null) {
            return;
        }
        try {
            LogCommand.on(this.repository).rev(getNode()).execute();
            if (this.data == null) {
                throw new IllegalStateException("data was not loaded");
            }
        } catch (IOException e) {
            throw new RuntimeIOException(e);
        }
    }

    @Override
    public boolean equals(Object that) {
        if (that instanceof Changeset) {
            return equals((Changeset) that);
        } else {
            return false;
        }
    }

    public boolean equals(Changeset that) {
        if (this.repository != that.repository) {
            return false;
        }
        return this.getNode().equals(that.getNode());
    }

    @Override
    public int hashCode() {
        return getRevision();
    }

    @Override
    public String toString() {
        return "changeset[" + this.node + "]";
    }

    @VisibleForTesting
    ChangesetData getData() {
        return this.data;
    }

    Repository getRepository() {
        return repository;
    }

}

class ChangesetData {
    public int revision;
    public String node;
    public String user;
    public DateTime timestamp;
    public String branch;
    public Changeset parent1;
    public Changeset parent2;
    public String message;

    public ChangesetData(int revision, String node, String user, DateTime timestamp, String branch, Changeset parent1,
            Changeset parent2, String message) {
        this.revision = revision;
        this.node = node;
        this.user = user;
        this.timestamp = timestamp;
        this.branch = branch;
        this.parent1 = parent1;
        this.parent2 = parent2;
        this.message = message;
    }

}
