/*
 * #%L
 * JavaHg
 * %%
 * Copyright (C) 2011 aragost Trifork ag
 * %%
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * #L%
 */
package com.aragost.javahg.commands;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import org.joda.time.DateTime;

import com.aragost.javahg.Changeset;
import com.aragost.javahg.Repository;
import com.aragost.javahg.internals.BlockInputStream;
import com.aragost.javahg.internals.HgInputStream;
import com.aragost.javahg.internals.OutputChannelInputStream;
import com.aragost.javahg.internals.RuntimeIOException;
import com.aragost.javahg.internals.Utils;
import com.google.common.collect.Lists;
import com.google.common.io.ByteStreams;

/**
 * Base class for the command classes.
 * 
 * Each Mercurial command (e.g., "log", "commit", etc) is mapped to a
 * command class, which is a subclass of this class. The command
 * classes will provide methods for setting command line flags and for
 * actually executing the command.
 */
public abstract class AbstractCommand {

    private final List<String> cmdLine;

    private final Repository repository;

    private OutputChannelInputStream outputChannelStream;

    private ByteArrayOutputStream error = new ByteArrayOutputStream();

    private int returnCode = Integer.MIN_VALUE;

    private int lineChannelLength;

    protected AbstractCommand(Repository repository) {
        this.repository = repository;
        this.cmdLine = Lists.newArrayList(getCommandName());
    }

    /**
     * @return the name of this Mercurial command, i.e., "add", "log",
     *         etc
     */
    public abstract String getCommandName();

    protected void cmdAppend(String option) {
        this.cmdLine.add(option);
    }

    protected void cmdAppend(String option, String arg) {
        if (arg == null) {
            throw new NullPointerException("cannot pass null for " + option + " flag");
        }
        this.cmdLine.add(option);
        this.cmdLine.add(arg);
    }

    protected void cmdAppend(String option, String[] args) {
        for (String arg : args) {
            cmdAppend(option, arg);
        }
    }

    protected void cmdAppend(String option, int arg) {
        this.cmdLine.add(option);
        this.cmdLine.add("" + arg);
    }

    protected void cmdAppend(String option, DateTime arg) {
        if (arg == null) {
            throw new NullPointerException("cannot pass null for " + option + " flag");
        }
        this.cmdLine.add(option);
        this.cmdLine.add(Utils.formatDate(arg));
    }

    /**
     * Launch the command and return stdout as a String.
     * 
     * @param args
     *            extra command line arguments (optional).
     * @return stdout as a String.
     * @throws IOException
     */
    protected final String launchString(String... args) throws IOException {
        InputStream stdout = launchStream(args);
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        byte[] bytes = new byte[1024];
        while (true) {
            int r = stdout.read(bytes);
            if (r == -1) {
                break;
            }
            out.write(bytes, 0, r);
        }
        ByteArrayInputStream in = new ByteArrayInputStream(out.toByteArray());
        try {
            return Utils.readStream(in, getRepository().getServer().getDecoder());
        } finally {
            cleanUp();
        }
    }

    /**
     * Launch the command and return stdout as a InputStream.
     * 
     * @param args
     *            extra command line arguments (optional).
     * @return stdout stream
     */
    protected final HgInputStream launchStream(String... args) {
        clear();
        List<String> commandLine = new ArrayList<String>(this.cmdLine);
        getRepository().addToCommandLine(commandLine);
        commandLine.addAll(Arrays.asList(args));
        try {
            this.outputChannelStream = this.repository.getServer().runCommand(commandLine, this);
            return new HgInputStream(outputChannelStream, this.repository.getServer().getDecoder());
        } catch (IOException e) {
            throw new RuntimeIOException(e);
        }
    }

    protected final LineIterator launchIterator(String... args) {
        return new LineIterator(launchStream(args));
    }

    /**
     * Open the output stream again after sending input to the command
     * server. When the server alternates between sending output on
     * the 'o' channel and reading input lines, the output channel
     * will run dry several times. Call this method after sending
     * input to the command server in response to reading from the 'L'
     * channel. New output from the server will then appear on the
     * output channel.
     */
    public void reopenOutputChannelStream() {
        this.outputChannelStream.reopen();
    }

    /**
     * Send input line to command server in response to reading a
     * block on the 'L' channel.
     * 
     * @param s
     *            line of input.
     */
    public void sendLine(String s) {
        if (this.lineChannelLength == 0) {
            throw new IllegalStateException("No input expected");
        }

        getRepository().getServer().sendLine(s);
        this.lineChannelLength = 0;
    }

    /**
     * @return true if we have read a 'L' channel from the server.
     */
    public boolean needsInputLine() {
        return this.lineChannelLength > 0;
    }

    /**
     * Finish the request by consuming any remaining input on the
     * output channel from the server. The server wont respond to new
     * commands until this is done.
     */
    public void cleanUp() {
        if (this.outputChannelStream != null) {
            try {
                Utils.consumeAll(this.outputChannelStream);
            } catch (IOException e) {
                throw new RuntimeIOException(e);
            }
        }
    }

    protected void clear() {
        this.error.reset();
        this.returnCode = Integer.MIN_VALUE;
    }

    /**
     * Check if the command ended with a zero return code. Subclasses
     * can override this to accept other return codes as successful.
     * 
     * @return true if the command ended successfully.
     */
    public boolean isSuccessful() {
        return getReturnCode() == 0;
    }

    protected String readInputStreamAsString(InputStream in) throws IOException {
        byte[] bytes = ByteStreams.toByteArray(in);
        return new String(bytes);
    }

    private String streamAsString(OutputStream stream, Charset charset) {
        if (stream instanceof ByteArrayOutputStream) {
            byte[] bytes = ((ByteArrayOutputStream) stream).toByteArray();
            try {
                CharsetDecoder decoder = getRepository().getServer().getDecoder();
                CharBuffer decoded = decoder.decode(ByteBuffer.wrap(bytes));
                return decoded.toString();
            } catch (CharacterCodingException ex) {
                throw new RuntimeException("Could not decode stream", ex);
            }
        } else {
            throw new IllegalStateException();
        }
    }

    /**
     * @return data read on the 'e' channel
     */
    public String getErrorString() {
        return streamAsString(this.error, Utils.getUtf8Charset());
    }

    /**
     * @return the return code read from the 'r' channel
     */
    public int getReturnCode() {
        if (this.returnCode == Integer.MIN_VALUE) {
            throw new IllegalStateException("cmdserver is still executing request");
        }
        return this.returnCode;
    }

    /**
     * @param rc
     *            the return code
     */
    // TODO: this ought not to be public, only
    // OutputChannelInputStream calls it and it should do so more
    // privately.
    public void setReturnCode(int rc) {
        this.returnCode = rc;
    }

    /**
     * @return the {@link Repository} associated with this command.
     */
    public Repository getRepository() {
        return repository;
    }

    @Override
    public String toString() {
        return getCommandName();
    }

    /**
     * @param lineChannelLength
     *            the line length requested by the server on the 'L'
     *            channel.
     */
    // TODO: as for setReturnCode, it would be nice if this was not
    // public.
    public void setLineChannelLength(int lineChannelLength) {
        this.lineChannelLength = lineChannelLength;
    }

    /**
     * Add to the stderr data stored in this command.
     * 
     * @param cin
     *            stderr of the command server.
     * @throws IOException
     */
    // TODO: as for setReturnCode, it would be nice if this was not
    // public.
    public void addToError(BlockInputStream cin) throws IOException {
        ByteStreams.copy(cin, this.error);
    }

    protected void withDebugAndChangesetTemplate() {
        cmdAppend("--debug");
        cmdAppend("--template", Changeset.CHANGESET_TEMPLATE);
    }

}

class LineIterator implements Iterator<String>, Iterable<String> {

    private boolean atEnd = false;

    private HgInputStream stream;

    public LineIterator(HgInputStream stream) {
        this.stream = stream;
    }

    public Iterator<String> iterator() {
        return this;
    }

    public boolean hasNext() {
        if (!this.atEnd) {
            try {
                this.atEnd = (this.stream.peek() == -1);
            } catch (IOException e) {
                throw new RuntimeIOException(e);
            }
        }
        return !this.atEnd;
    }

    public String next() {
        if (hasNext()) {
            try {
                return this.stream.textUpTo('\n');
            } catch (IOException e) {
                throw new RuntimeIOException(e);
            }
        } else {
            throw new NoSuchElementException();
        }
    }

    public void remove() {
        throw new UnsupportedOperationException();
    }

}
