/*
 * #%L
 * JavaHg
 * %%
 * Copyright (C) 2011 aragost Trifork ag
 * %%
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * #L%
 */
package com.aragost.javahg.commands;

import java.io.IOException;
import java.util.List;

import com.aragost.javahg.Repository;
import com.aragost.javahg.commands.flags.AddCommandFlags;
import com.google.common.collect.Lists;

/**
 * Command class for executing <tt>hg add</tt>. Set flags from
 * {@link AddCommandFlags} and see the {@link #execute(String...)}
 * method for how to run the command.
 */
public class AddCommand extends AddCommandFlags {

    private static final String PREFIX = "adding ";

    /**
     * @param repository
     *            the repository associated with this command.
     */
    public AddCommand(Repository repository) {
        super(repository);
        cmdAppend("--debug");
    }

    /**
     * Check if the command ended successfully.
     * <p>
     * In contrast with Mercurial, a return code of 1 is considered to
     * be successful. Mercurial returns 1 if some files could not be
     * found, but at the same time it does add the files that could be
     * found.
     * 
     * @return true if the command exited with 0 or 1.
     */
    @Override
    public boolean isSuccessful() {
        return super.isSuccessful() || getReturnCode() == 1;
    }

    /**
     * Execute the add command, and return list of files added.
     * <p>
     * If a file could not be found, then Mercurial adds the other
     * files and exit with a return code of 1. As described in
     * {@link #isSuccessful()}, we don't treat that as an error and
     * the caller should instead inspect the returned list to see if
     * all files were added.
     * 
     * @param paths
     *            the files or directories to add. Leave empty to add
     *            all untracked files.
     * @return list of files added actually.
     * @throws IOException
     */
    public List<String> execute(String... paths) throws IOException {
        List<String> result = Lists.newArrayList();
        for (String line : launchIterator(paths)) {
            if (line.startsWith(PREFIX)) {
                String file = new String(line.substring(PREFIX.length()));
                result.add(file);
            } else {
                cleanUp();
                throw new IllegalStateException("Unexpected line in output from add command");
            }
        }
        return result;
    }

}
