/*
 * #%L
 * JavaHg
 * %%
 * Copyright (C) 2011 aragost Trifork ag
 * %%
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * #L%
 */
package com.aragost.javahg.commands;

import java.io.IOException;
import java.util.List;
import java.util.Map;

import com.aragost.javahg.Changeset;
import com.aragost.javahg.Repository;
import com.aragost.javahg.commands.flags.AnnotateCommandFlags;
import com.aragost.javahg.internals.HgInputStream;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

/**
 * Command class for executing <tt>hg annotate</tt>. Set flags from
 * {@link AnnotateCommandFlags} and see the {@link #execute} method
 * for how to run the command.
 */
public class AnnotateCommand extends AnnotateCommandFlags {

    /**
     * @param repository
     *            the repository associated with this command.
     */
    public AnnotateCommand(Repository repository) {
        super(repository);
        cmdAppend("--number");
    }

    /**
     * Execute <tt>hg annotate</tt> for the specified file.
     * <p>
     * On the command line, Mercurial can annotate more than one file,
     * JavaHg can only annotate one file for each call (but the
     * overhead is negligible to make multiple calls).
     * 
     * @param file
     * @return list of annotated lines.
     * @throws IOException
     */
    public List<AnnotateLine> execute(String file) throws IOException {
        Repository repo = getRepository();
        HgInputStream stream = launchStream(file);
        LogCommand logCmd = LogCommand.on(repo);
        Map<Integer, String> dataMap = Maps.newHashMap();
        List<AnnotateLine> result = Lists.newArrayList();
        while (stream.peek() != -1) {
            Integer rev = stream.decimalIntUpTo(':');
            logCmd.rev(rev.toString());
            stream.skip(1);
            String data = stream.textUpTo('\n');
            dataMap.put(rev, data);
        }
        List<Changeset> changesets = logCmd.execute();
        for (Changeset changeset : changesets) {
            String data = dataMap.get(changeset.getRevision());
            result.add(new AnnotateLine(changeset, data));
        }

        return result;
    }

}
