/*
 * #%L
 * JavaHg
 * %%
 * Copyright (C) 2011 aragost Trifork ag
 * %%
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * #L%
 */
package com.aragost.javahg.internals;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.aragost.javahg.commands.AbstractCommand;
import com.aragost.javahg.commands.ExecutionException;

/**
 * An input stream that will make everything written to the output
 * channel for one command available as an input stream.
 * <p>
 * Reading from this stream will add to the
 * {@link ByteArrayOutputStream} for the other channels, as block for
 * other channels are detected.
 * <p>
 * The stream will indicate EOF when the 'r'esult block has been read.
 */
public class OutputChannelInputStream extends InputStream {

    private static final Logger LOG = LoggerFactory.getLogger(OutputChannelInputStream.class);

    private InputStream channelStream;

    private AbstractCommand cmd;

    private BlockInputStream currentOutputChannelBlock;

    private boolean eof = false;

    /**
     * 
     * @param channelStream
     * @param cmd
     * @throws IOException
     */
    public OutputChannelInputStream(InputStream channelStream, AbstractCommand cmd) throws IOException {
        this.channelStream = channelStream;
        this.cmd = cmd;
        findNextOutputChannelBlock();
    }

    @Override
    public int read() throws IOException {
        if (this.eof) {
            return -1;
        }
        int ch = this.currentOutputChannelBlock.read();
        if (ch == -1) {
            findNextOutputChannelBlock();
            return read();
        }
        return ch;
    }

    @Override
    public int read(byte[] b, int off, int len) throws IOException {
        if (this.eof) {
            return -1;
        }
        int n = this.currentOutputChannelBlock.read(b, off, len);
        if (n > 0) {
            return n;
        } else if (n == -1) {
            findNextOutputChannelBlock();
            return read(b, off, len);
        } else {
            return 0;
        }
    }

    /**
     * Find the next output channel block.
     * <p>
     * The method will handle other channels while looking for an
     * o-channel block.
     * 
     * @throws IOException
     */
    private void findNextOutputChannelBlock() throws IOException {
        if (this.currentOutputChannelBlock != null) {
            if (this.currentOutputChannelBlock.available() > 0) {
                throw new IllegalStateException();
            }
        }
        if (this.eof) {
            throw new IllegalStateException();
        }
        BlockInputStream cin = null;
        loop: while (true) {
            cin = new BlockInputStream(channelStream);
            char channel = cin.getChannel();
            switch (channel) {
            case 'o':
                break loop;
            case 'e':
                this.cmd.addToError(cin);
                break;
            case 'r':
                if (cin.available() != 4) {
                    throw new IllegalStateException("For channel 'r' is length 4 expected");
                }
                int returnCode = cin.readInt();
                this.cmd.setReturnCode(returnCode);
                LOG.info("Return code: {}", returnCode);
                this.eof = true;
                this.cmd.getRepository().getServer().clearCurrentCommand(this.cmd);
                if (!this.cmd.isSuccessful()) {
                    throw new ExecutionException(cmd);
                }
                break loop;
            case 'L':
                this.cmd.setLineChannelLength(cin.getLength());
                this.eof = true;
                break loop;
            default:
                if (Character.isUpperCase(channel)) {
                    throw new IllegalStateException("Unknown channel: " + channel);
                } else {
                    cin.skip();
                }
            }
        }
        this.currentOutputChannelBlock = cin;
    }

    public void reopen() {
        this.eof = false;
        try {
            findNextOutputChannelBlock();
        } catch (IOException e) {
            throw new RuntimeIOException(e);
        }
    }
}
