/*
 * #%L
 * JavaHg
 * %%
 * Copyright (C) 2011 aragost Trifork ag
 * %%
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * #L%
 */
package com.aragost.javahg.internals;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CodingErrorAction;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.aragost.javahg.commands.AbstractCommand;
import com.google.common.collect.Lists;

public class Server {

    private static final Logger LOG = LoggerFactory.getLogger(Server.class);

    /**
     * If a lock can't be obtained with in this number of seconds a
     * warning is written to the log
     */
    private static final int LOCK_WAIT_WARN = 10;

    /**
     * Fail (i.e. throw an exception) if a lock can't be obtained
     * within this number of seconds
     */
    private static final int LOCK_WAIT_FAIL = 120;

    private final Charset encoding = Utils.getUtf8Charset();

    /**
     * Policy for handling decoding errors.
     */
    private CodingErrorAction errorAction = CodingErrorAction.REPORT;

    /**
     * The underlying OS process
     */
    private Process process;

    /**
     * A server can be used by more than one
     * {@link com.aragost.javahg.Repository}. This is a reference
     * count for repository. When it reached zero the server is
     * stopped
     */
    private int refCount;

    private Lock lock = new ReentrantLock();

    /**
     * Thread that read stderr from command server
     */
    private Thread errorReaderThread;

    /**
     * If non-null the {@link AbstractCommand} the server is executing
     */
    private AbstractCommand currentCommand;

    /**
     * The directory containing the Mercurial repository where the
     * server is running.
     */
    private File directory;

    private final String hgBin;

    public Server(String hgBin) {
        this.hgBin = hgBin;
    }

    public CodingErrorAction getErrorAction() {
        return errorAction;
    }

    public void setErrorAction(CodingErrorAction errorAction) {
        this.errorAction = errorAction;
    }

    public CharsetDecoder getDecoder() {
        CharsetDecoder decoder = this.encoding.newDecoder();
        decoder.onMalformedInput(this.errorAction);
        decoder.onUnmappableCharacter(this.errorAction);
        return decoder;
    }

    public CharsetEncoder getEncoder() {
        CharsetEncoder encoder = this.encoding.newEncoder();
        encoder.onMalformedInput(this.errorAction);
        encoder.onUnmappableCharacter(this.errorAction);
        return encoder;
    }

    /**
     * Encode the string and write it to the OutputStream
     * 
     * @param s
     * @param output
     * @throws IOException
     */
    public void encode(String s, OutputStream output) throws IOException {
        CharsetEncoder encoder = getEncoder();
        ByteBuffer byteBuffer = encoder.encode(CharBuffer.wrap(s));
        output.write(byteBuffer.array(), byteBuffer.arrayOffset(), byteBuffer.limit());
    }

    /**
     * Start the server in the specified directory. The directory is
     * assumed to be a Mercurial repository.
     * 
     * @param directory
     * @return the hello message from the server
     */
    public String start(File directory) {
        this.directory = directory.getAbsoluteFile();
        if (!new File(directory, ".hg").isDirectory()) {
            throw new IllegalStateException("No .hg in " + directory);
        }
        try {
            this.process = execHg(this.directory, "serve", "--cmdserver", "pipe", "--config", "ui.interactive=true");
            BlockInputStream cin = new BlockInputStream(this.process.getInputStream());
            Runnable errorReader = new Runnable() {
                public void run() {
                    BufferedReader reader = new BufferedReader(new InputStreamReader(
                            Server.this.process.getErrorStream()));
                    String line;
                    try {
                        while ((line = reader.readLine()) != null) {
                            System.err.println(line);
                        }
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            };
            this.errorReaderThread = new Thread(errorReader);
            errorReaderThread.start();
            return Utils.readStream(cin, getDecoder());
        } catch (IOException e) {
            throw new RuntimeIOException(e);
        }
    }

    public File getDirectory() {
        return directory;
    }

    public void stop() {
        if (this.process == null) {
            throw new IllegalArgumentException("Server is not running");
        }
        closeStream(this.process.getOutputStream());
        closeStream(this.process.getInputStream());
        try {
            this.errorReaderThread.join(1000);
        } catch (InterruptedException e) {
            // TODO handle this better
            this.errorReaderThread = null;
        }
        closeStream(this.process.getErrorStream());
        this.process.destroy();
        this.currentCommand = null;
        this.process = null;
        this.directory = null;
    }

    private void closeStream(Closeable stream) {
        try {
            stream.close();
        } catch (IOException e) {
            // Not so much we can do :-(
            e.printStackTrace();
        }
    }

    /**
     * Run the specified command and return a stream with the content
     * of the output channel.
     * <p>
     * The client <em>must</em> empty the return stream. The server
     * will not accept other commands until all out is read.
     * 
     * @param cmdLine
     * @param command
     * @return the standard output from the command.
     * @throws IOException
     */
    public OutputChannelInputStream runCommand(List<String> cmdLine, AbstractCommand command) throws IOException {
        try {
            if (!this.lock.tryLock(LOCK_WAIT_WARN, TimeUnit.SECONDS)) {
                LOG.warn("Waited " + LOCK_WAIT_WARN + " seconds for server lock without obtaining it");
                if (!this.lock.tryLock(LOCK_WAIT_FAIL, TimeUnit.SECONDS)) {
                    String msg = "Did not obtain server lock after " + LOCK_WAIT_FAIL + " seconds.";
                    LOG.error(msg);
                    throw new RuntimeException(msg);
                }
            }
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
        if (this.currentCommand != null) {
            throw new IllegalStateException("Trying to execute new command when command already running: "
                    + this.currentCommand);
        }
        this.currentCommand = command;
        sendCommand(cmdLine);
        OutputChannelInputStream stdout = new OutputChannelInputStream(this.process.getInputStream(), command);

        return stdout;
    }

    public void clearCurrentCommand(AbstractCommand cmd) {
        if (cmd != this.currentCommand) {
            throw new IllegalStateException("Wrong command");
        }
        this.currentCommand = null;
        this.lock.unlock();
    }

    private void sendCommand(List<String> cmdLine) throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        encode(cmdLine.get(0), baos);
        for (String s : cmdLine.subList(1, cmdLine.size())) {
            baos.write('\0');
            encode(s, baos);
        }

        OutputStream outputStream = this.process.getOutputStream();
        outputStream.write("runcommand\n".getBytes());
        Utils.writeBigEndian(baos.size(), outputStream);
        baos.writeTo(outputStream);
        outputStream.flush();
    }

    private Process execHg(File directory, String... arguments) throws IOException {
        ProcessBuilder processBuilder = new ProcessBuilder(Lists.asList(this.hgBin, arguments));
        if (directory != null) {
            processBuilder.directory(directory);
        }
        Map<String, String> environment = processBuilder.environment();
        environment.put("HGENCODING", this.encoding.displayName());
        environment.put("HGPLAIN", "1");
        environment.put("HGRCPATH", "");
        return processBuilder.start();
    }

    private String readStream(InputStream in) {
        return Utils.readStream(in, getDecoder());
    }

    /**
     * Convenience method to initialize a mercurial repository in a
     * directory.
     * <p>
     * This method is not using any commandserver functionality
     * 
     * @param directory
     */
    public void initMecurialRepository(File directory) {
        execHgCommand(null, "init", directory.getAbsolutePath());
    }

    public void execHgCommand(File directory, String... args) {
        try {
            Process process = execHg(directory, args);
            String stderr = readStream(process.getErrorStream());
            readStream(process.getInputStream());

            if (process.waitFor() != 0) {
                throw new RuntimeException(stderr);
            }
        } catch (IOException e) {
            throw new RuntimeIOException(e);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public String toString() {
        return "cmdserver@" + this.directory;
    }

    @Override
    protected void finalize() {
        if (this.process != null) {
            LOG.error("Stopping cmdserver via finalize. Please explicit stop server.");
            stop();
        }
    }

    public void sendLine(String answer) {
        OutputStream outputStream = this.process.getOutputStream();
        try {
            byte[] bytes = (answer + "\n").getBytes(this.encoding.name());
            Utils.writeBigEndian(bytes.length, outputStream);
            outputStream.write(bytes);
            outputStream.flush();
        } catch (IOException e) {
            throw new RuntimeIOException(e);
        }
    }

    public Process getProcess() {
        return process;
    }

    public void incrementRefCount() {
        this.refCount++;
    }

    public void decrementRefCount() {
        this.refCount--;
        if (this.refCount == 0) {
            stop();
        }
    }
}
